package gov.fnal.xpathsearch.cache.filesystem;

import org.w3c.dom.Document;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.*;

import java.io.*;
import java.util.HashMap;
import java.util.LinkedHashMap;

import gov.fnal.xpath2ldap.XMLAdapter;
import gov.fnal.xpath2ldap.util.ElementAttributeMergeConfigurer;
import gov.fnal.xpath2ldap.util.PrettyTransform;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.ParserConfigurationException;

import ch.elca.el4j.services.xmlmerge.XmlMerge;
import ch.elca.el4j.services.xmlmerge.config.ConfigurableXmlMerge;

public class AccessFilesystem implements URIResolver {

    private static HashMap<String, File> filehash = new HashMap();
    protected static TransformerFactory tFactory;
    protected static DocumentBuilder dBuilder;
    protected static Transformer merge_transformer;
    protected static LinkedHashMap<String, Document> mergeDocuments = new LinkedHashMap<String, Document>();



    public static void create(Document doc_in, String docName, String docDir) throws Exception {
        create(doc_in, docName, docDir, true, true);
    }

    public static void create(Document doc_in, String docName, String docDir, boolean sort, boolean pretty) throws Exception {

        File xmlfile;
        boolean completed = false;
        try {
            xmlfile = getFile(docName, docDir);
            try {
                FileOutputStream fos = new FileOutputStream(xmlfile);
                Document sorted_doc;
                if (sort) {
                    sorted_doc  = getSorted(doc_in);
                } else {
                    sorted_doc = doc_in;
                }
                Document pretty_doc;
                if (pretty) {
                    pretty_doc = getPretty(sorted_doc);
                } else {
                    pretty_doc = sorted_doc;
                }
                XMLAdapter.serialize(pretty_doc, fos);
                fos.flush();
                fos.close();
            } catch (Exception e) {
                System.err.println("Could not serialize document " + e);
            }
            //try {
            //    xmlfile.setWritable(true, false);
            //} catch (Exception e) {
            //    System.err.println("Could not make document world-writeable" + e);
            //}
            completed = true;
        } catch (Exception e) {
            System.err.println("Could not create document.");
            e.printStackTrace();
            //rollback
        } //finally {
        //  closeFile(xmlfile);
        //}
    }

    public static Document read(String docName, String docDir) throws Exception {
        File xmlfile;
        FileInputStream fis=null;
        Document doc_out=null;
        boolean completed = false;
        try {
            xmlfile = getFile(docName, docDir);
            if(!xmlfile.exists()) return null;
            fis = new FileInputStream(xmlfile);
            InputSource source = new InputSource(fis);
            DocumentBuilderFactory dbf = null;
            DocumentBuilder db = null;
            try {
                if ( dbf == null ) {
                    dbf = DocumentBuilderFactory.newInstance();
                    dbf.setNamespaceAware( true );
                    dbf.setValidating( false );
                }
            } catch ( Exception e ) {
                e.printStackTrace();
            }
            db = dbf.newDocumentBuilder();
            doc_out = db.parse( source );
            completed = true;
        } catch (Exception e) {
            System.err.println("Could not update document.");
            e.printStackTrace();
            //rollback
        } finally {
            if(fis!=null) {
                fis.close();
            }
        }
        return doc_out;
    }

    public static Document update(Document from_doc, String docName, String docDir) throws Exception {
        if(from_doc==null) return read(docName, docDir);
        Document merged_doc = null;
        boolean completed = false;
        try {
            merged_doc = merge(from_doc, docName, docDir);
            delete(docName, docDir);//TODO Keep previous versions.
            create(merged_doc, docName, docDir, false, true);
            completed = true;
        } catch (Exception e) {
            System.err.println("Could not update document.");
            e.printStackTrace();
            //rollback
        }
        return merged_doc;
    }

    public static void delete(String docName, String docDir) throws Exception {
        File xmlfile;
        boolean completed = false;
        try {
            xmlfile = getFile(docName, docDir);
            if(xmlfile.exists()) {
                xmlfile.delete();
            }
            completed = true;
        } catch (Exception e) {
                System.err.println("Could not delete document.");
                e.printStackTrace();
          //rollback
        } //finally {
          //  closeFile(xmlfile);
        //}
    }

    public static File getFile(String docName, String dbName) {
        File cacheDir = new File(dbName);
        if(!cacheDir.isDirectory()) {
            System.err.println(dbName + " is not a directory.\nPlease create the directory which is to hold the cache files.");
        }
        if(!cacheDir.canWrite()) {
            System.err.println(dbName + " is not writeable.\nPlease make writable the directory which is to hold the cache files.");
        }
        String docPath = dbName +  File.separator + docName;
        File xmlfile = new File(docPath);
        return xmlfile;
    }

    public static Document merge(Document from_doc, String docName, String docDir) throws Exception {
        Document into_doc = read(docName, docDir);
        if (from_doc==null) return into_doc;
        // Note: both documents must be pre-sorted by the attribute specified in ElementAttributeMergeConfigurer
        Document sorted_from_doc  = ElementAttributeMergeConfigurer.getSorted(from_doc);
        if (into_doc==null) return sorted_from_doc;
        XmlMerge xmlMerge = new ConfigurableXmlMerge(new ElementAttributeMergeConfigurer());
        final Document[] sources = new Document[]{sorted_from_doc, into_doc};
        Document merged = xmlMerge.merge(sources);
        return merged;
    }

    public static Document getSorted(Document doc_in) throws Exception {
        return ElementAttributeMergeConfigurer.getSorted(doc_in);
    }

    public static Document makeSorted(Document doc_in, String docName, String docDir) throws Exception {
        if(doc_in==null) return doc_in;
        Document sorted_doc = null;
        boolean completed = false;
        try {
            sorted_doc = getSorted(doc_in);
            delete(docName, docDir);//TODO Keep previous versions.
            create(sorted_doc, docName, docDir);
            completed = true;
        } catch (Exception e) {
            System.err.println("Could not update document.");
            e.printStackTrace();
            //rollback
        }
        return sorted_doc;
    }

    public static Document getPretty(Document doc_in) throws Exception {
        if(doc_in==null) return doc_in;
        Document pretty_doc = PrettyTransform.doPrettyTransform(doc_in);
        return pretty_doc;
    }

    public static Document makePretty(Document doc_in, String docName, String docDir) throws Exception {
        if(doc_in==null) return doc_in;
        Document pretty_doc = null;
        boolean completed = false;
        try {
            pretty_doc = getPretty(doc_in);
            delete(docName, docDir);//TODO Keep previous versions.
            create(pretty_doc, docName, docDir);
            completed = true;
        } catch (Exception e) {
            System.err.println("Could not update document.");
            e.printStackTrace();
            //rollback
        }
        return pretty_doc;
    }

    public static Document merge3(Document from_doc, Document into_doc) {
        if (into_doc==null) return from_doc;
        if (from_doc==null) return into_doc;

        DOMSource searchSource = new DOMSource(from_doc);
        DOMResult mergeResult = new DOMResult();
        try {
            //Document merge_xsl = dBuilder.parse("/home/tdh/util/schematools-0.0.1/merge.xslt");
            //DOMSource merge_xslSource = new DOMSource(merge_xsl);
            //merge_xslSource.setSystemId("merge.xslt");
            //Transformer func_merge_transformer = tFactory.newTransformer(merge_xslSource);
            //func_merge_transformer.setURIResolver(this);
            mergeDocuments.put("mergefuncdoc", into_doc);
            getMergeTransformer().setParameter("with", "mergefuncdoc");
            //String xformed_search_result_str = XMLAdapter.testDOM3LS(from_doc);
                //System.out.println(xformed_search_result_str);
            //Document better_xformed_search_result = dBuilder.parse(new InputSource(new StringReader(xformed_search_result_str)));
            //searchSource = new DOMSource(better_xformed_search_result);

            //StringWriter xform_result = new StringWriter();
            //merge_transformer.transform(searchSource, new StreamResult(xform_result));
            getMergeTransformer().transform(searchSource, mergeResult);
            //String into_doc_str = XPathLDAPSearch.testDOM3LS((Document)mergeResult.getNode());
            //String into_doc_str = xform_result.toString();
            //System.out.println(into_doc_str);
            //return dBuilder.parse(new InputSource(new StringReader(into_doc_str)));
            return (Document)mergeResult.getNode();
        //} catch (IOException pce) {
        //    System.err.println("IOException");
        } catch (TransformerException pce) {
            System.err.println("TransformerException");
        //} catch (SAXException pce) {
        //    System.err.println("SAXException");
        } catch (Exception e) {
            System.err.println("Exception");
        }
        return null;
    }

    public static Transformer getMergeTransformer() {
        if(merge_transformer==null) {
            tFactory = TransformerFactory.newInstance();
            if(tFactory.getFeature(DOMSource.FEATURE) && tFactory.getFeature(DOMResult.FEATURE))
            {
                try {
                    //Instantiate a DocumentBuilderFactory.
                    System.setProperty("javax.xml.parsers.XmlDocumentParserFactory", "org.apache.xerces.jaxp.DocumentBuilderFactoryImpl");
                    DocumentBuilderFactory dFactory = DocumentBuilderFactory.newInstance();

                    // And setNamespaceAware, which is required when parsing xsl files
                    dFactory.setNamespaceAware(true);

                    //Use the DocumentBuilderFactory to create a DocumentBuilder.
                    dBuilder = dFactory.newDocumentBuilder();
                    Document merge_xsl = dBuilder.parse("/home/tdh/util/schematools-0.0.1/merge4.xslt");//TODO relative path
                    DOMSource merge_xslSource = new DOMSource(merge_xsl);
                    merge_xslSource.setSystemId("merge.xslt");
                    merge_transformer = tFactory.newTransformer(merge_xslSource);
                    merge_transformer.setURIResolver(new AccessFilesystem());
                } catch (ParserConfigurationException pce) {
                    System.err.println("ParserConfigurationException");
                } catch (IOException pce) {
                    System.err.println("IOException");
                } catch (TransformerException pce) {
                    System.err.println("TransformerException");
                } catch (SAXException pce) {
                    System.err.println("SAXException");
                }
            }  else
            {
                //throw new org.xml.sax.SAXNotSupportedException("DOM node processing not supported!");
                System.err.println("DOM node processing not supported!");
            }
        }

        return merge_transformer;
    }

    public Source resolve(String publicId, String systemId)
    {
        //System.out.println("resolving " + publicId + " and  " + systemId);
        if (true)
        {
            //try {
            //    DocumentBuilderFactory dFactory = DocumentBuilderFactory.newInstance();

            // And setNamespaceAware, which is required when parsing xsl files
            //dFactory.setNamespaceAware(true);

            //Use the DocumentBuilderFactory to create a DocumentBuilder.
            //DocumentBuilder dBuilder = dFactory.newDocumentBuilder();
            //Document better_attrsearchdoc = dBuilder.parse(new InputSource(new StringReader(attrsearchdoc_str)));
            return new DOMSource(mergeDocuments.get(publicId));
            //} catch (Exception e) {
            //    System.err.println("Could not parse attrsearchdoc");
            //}
        }

        return null;
    }
}