xquery version "1.0";
(: Returns a list of example srm-sp-reserve commands for Storage Elements on which the specified VO is authorized. :)

import module namespace osg="http://opensciencegrid.org" at "functions.xq";

declare variable $vo as xs:string external;
declare variable $fqan := osg:clean_fqan($vo);
declare variable $description as xs:string external;
declare variable $requested_size as xs:string external;
declare variable $guaranteed_size as xs:string external;
declare variable $lifetime as xs:string external;
declare variable $storage_element_id as xs:string external;
declare variable $free_online_size as xs:decimal external;
declare variable $show_storage_element_id as xs:string external;

declare variable $description_arg := if ($description="") then '' else concat(" -space_desc=", $description);
declare variable $requested_size_arg := if ($requested_size="") then '' else concat(" -desired_size=", $requested_size);
declare variable $guaranteed_size_arg := if ($guaranteed_size="") then '' else concat(" -guaranteed_size=", $guaranteed_size);
declare variable $lifetime_arg := if ($lifetime="") then '' else concat(" -lifetime=", $lifetime);
declare variable $free_online_size_arg as xs:decimal:= if ($free_online_size=0) then 0 else $free_online_size;
declare variable $header_str := concat("&#x22;", "STORAGE ELEMENT ID", "&#x22;", "&#x20;", "&#x22;", "SRM_SP_RESERVE COMMAND", "&#x22;", "&#xA;");
declare variable $header := if ($show_storage_element_id='') then $header_str else "";

concat($header,
string-join(
for $acbr in (//GlueSE[@GlueSEUniqueID]/GlueSEControlProtocol[@GlueSEControlProtocolVersion='2.2.0'][@GlueSEControlProtocolEndpoint]/../GlueSA[@GlueSAFreeOnlineSize>=$free_online_size_arg]/(@GlueSAAccessControlBaseRule | GlueVOInfo/@GlueVOInfoAccessControlBaseRule))
  let $child_acbrs := $acbr/../GlueVOInfo[@GlueVOInfoPath]/@GlueVOInfoAccessControlBaseRule
  let $endpoint := $acbr/(.|..)/../..[@GlueSEUniqueID]/GlueSEControlProtocol[@GlueSEControlProtocolVersion='2.2.0']/@GlueSEControlProtocolEndpoint
  let $seid := $acbr/(.|..)/../../@GlueSEUniqueID
    return
      if ($storage_element_id="" or $storage_element_id=$seid) then
        if(osg:acbr_match($acbr, $fqan)) then
          if($acbr/..[@GlueVOInfoPath] or ($acbr/..[@GlueSAPath] and fn:not(osg:acbr_nodes_match($child_acbrs, $fqan)))) then
	    if($show_storage_element_id='') then	  	        
	      concat("&#x22;", $seid, "&#x22;", "&#x20;", "&#x22;", "srm-sp-reserve", " -serviceurl", " srm", substring($endpoint[1], 6), $requested_size_arg, $guaranteed_size_arg, $lifetime_arg, $description_arg, "&#x22;", "&#xA;")
	    else
	      concat("&#x22;", "srm-sp-reserve", " -serviceurl", " srm", substring($endpoint[1], 6), $requested_size_arg, $guaranteed_size_arg, $lifetime_arg, $description_arg, "&#x22;", "&#xA;")
	  else()
	else()
      else() 
,""))
