package gov.fnal.xpath2ldap.util;

import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerException;
import javax.xml.transform.Transformer;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.dom.DOMResult;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.DocumentBuilder;
import java.net.URL;
import java.io.IOException;

import org.w3c.dom.Document;
import org.xml.sax.SAXException;


public class DnSortTransform {

    private static TransformerFactory tFactory;
    private static Transformer dnSortXform;

    public static Transformer getDnSortXform() {

        if(dnSortXform==null) {
            tFactory = TransformerFactory.newInstance();

            if(tFactory.getFeature(DOMSource.FEATURE) && tFactory.getFeature(DOMResult.FEATURE))
            {
                try {
                    //Instantiate a DocumentBuilderFactory.
                    System.setProperty("javax.xml.parsers.XmlDocumentParserFactory", "org.apache.xerces.jaxp.DocumentBuilderFactoryImpl");
                    DocumentBuilderFactory dFactory = DocumentBuilderFactory.newInstance();

                    // And setNamespaceAware, which is required when parsing xsl files
                    dFactory.setNamespaceAware(true);

                    //Use the DocumentBuilderFactory to create a DocumentBuilder.
                    DocumentBuilder dBuilder = dFactory.newDocumentBuilder();

                    //Use the DocumentBuilder to parse the XSL stylesheet.
                    //todo: remove hard-coded xsl transform name
                    URL dnSortXsl = DnSortTransform.class.getResource("/dn_sort.xsl");
                    Document dnSortDoc = dBuilder.parse(dnSortXsl.openStream());

                    // Use the DOM Document to define a DOMSource object.
                    DOMSource xslDomSource = new DOMSource(dnSortDoc);

                    // Set the systemId: note this is actually a URL, not a local filename
                    xslDomSource.setSystemId("dn_sort.xsl");

                    // Process the stylesheet DOMSource and generate a Transformer.
                    dnSortXform = tFactory.newTransformer(xslDomSource);

                } catch (ParserConfigurationException pce) {
                    System.err.println("ParserConfigurationException");
                } catch (IOException pce) {
                    System.err.println("IOException");
                } catch (TransformerException pce) {
                    System.err.println("TransformerException");
                } catch (SAXException pce) {
                    System.err.println("SAXException");
                }
            }  else
            {
                System.err.println("DOM node processing not supported!");
            }
        }

        return dnSortXform;
    }


    public static Document doDnSortTransform(Document unsorted) {
        // Transform the document by the dn attribute
        DOMSource xmlDomSource = new DOMSource(unsorted);
        xmlDomSource.setSystemId("unsortedDoc");
        DOMResult domResult = new DOMResult();
        try {
            getDnSortXform().transform(xmlDomSource, domResult);
        } catch (
                TransformerException pce) {
            System.err.println("TransformerException");
        }

        return (Document) domResult.getNode();
    }
}
