package gov.fnal.xpathsearch.cache;

import java.util.*;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

import org.apache.commons.cli.CommandLine;
import org.w3c.dom.Document;

public class XPathCacheLoad {

    static String expr1 ="//GlueCE/@GlueCEUniqueID";
    static String expr2 ="//GlueSE/GlueSEControlProtocol[@GlueSEControlProtocolVersion='2.2.0']";
    static String expr3 ="//GlueSE/GlueSA/@GlueSAAccessControlBaseRule";
    static String expr4 ="//GlueSE/GlueSA/@GlueSAFreeOnlineSize";
    static String expr5 ="//GlueCluster/@GlueClusterUniqueID";
    static String expr6 ="//GlueSE/GlueSEControlProtocol/@GlueSEControlProtocolEndpoint";
    static String expr7 ="//GlueCESEBind/@dn";
    static String expr8 ="//GlueCESEBind/@GlueCESEBindMountInfo";
    static String expr9 ="//GlueCluster/GlueSubCluster/@GlueChunkKey";
    static String expr10 ="//GlueCE/@GlueForeignKey";
    static String expr11 ="//GlueCEGlueVOView";
    static String expr12 ="//GlueSE/GlueSEAccessProtocol[@GlueSEAccessProtocolType='file']";
    static String expr13 ="//GlueSE/GlueSA/@GlueSAPath";
    static String expr14 ="//GlueCE/@GlueCEAccessControlBaseRule";
    static String expr15 ="//GlueSE/GlueSA/GlueVOInfo/@GlueVOInfoPath";
    static String expr16 ="//GlueCluster/GlueSubCluster/@GlueHostApplicationSoftwareRunTimeEnvironment";
    static String expr17 ="//GlueCluster/@GlueForeignKey";
    static String expr18 ="//GlueSE/GlueSA/GlueVOInfo/@GlueVOInfoAccessControlBaseRule";
    static String expr19 ="//GlueSE/@GlueSEUniqueID";


    public static Pattern EQ_SUBST = Pattern. compile("\\[(@.*=.*)\\]$");
    public static Pattern NE_SUBST = Pattern. compile("\\[(@.*!=.*)\\]$");
    public static Pattern LT_SUBST = Pattern. compile("\\[(@.*<.*)\\]$");
    public static Pattern LTE_SUBST = Pattern. compile("\\[(@.*<=.*)\\]$");
    public static Pattern GT_SUBST = Pattern. compile("\\[(@.*>.*)\\]$");
    public static Pattern GTE_SUBST = Pattern. compile("\\[(@.*>=.*)\\]$");
    public static Pattern[] ATTR_SUBST = new Pattern[]{EQ_SUBST, NE_SUBST, LT_SUBST, LTE_SUBST, GT_SUBST, GTE_SUBST};

    public static void main (String[] args) {
        if(args.length==0)  {
            args = new String[]{"--help"};
        }
        CommandLine cmdline = XPathCacheLoadOptions.processArguments(args);
        AccessCache.cache_type = cmdline.getOptionValue(XPathCacheLoadOptions.CACHE_TYPE);
        search_to_cache(cmdline);
    }

    public static void search_to_cache (CommandLine cmdline) {
        XPathCacheLoadEvaluator lpathSearcher = new XPathCacheLoadEvaluator(cmdline);
        List<String> lFullPaths = getFullPaths(cmdline);
        Document response_structured = lpathSearcher.search(lFullPaths);
        if(cmdline.hasOption("update")) {
            update_doc(response_structured);
        } else {
            create_doc(response_structured);
        }
    }

    private static void create_doc(Document document) {
        AccessCache.Operation.CREATE.go(document);
    }

    private static void update_doc(Document document) {
        AccessCache.Operation.UPDATE.go(document);
    }

    private static List<String> getFullPaths(CommandLine cmdline) {
        List<String> lFullPaths = new ArrayList<String>();
        List<String> argslist;

        if(cmdline.hasOption("test")) {
            argslist = new ArrayList<String>();
            argslist.add(expr1);
            argslist.add(expr2);
            argslist.add(expr3);
            argslist.add(expr4);
            argslist.add(expr5);
            argslist.add(expr6);
            argslist.add(expr7);
            argslist.add(expr8);
            argslist.add(expr9);
            argslist.add(expr10);
            argslist.add(expr11);
            argslist.add(expr12);
            argslist.add(expr13);
            argslist.add(expr14);
            argslist.add(expr15);
            argslist.add(expr16);
            argslist.add(expr17);
            argslist.add(expr18);
            argslist.add(expr19);
        } else {
            argslist = cmdline.getArgList();
        }

        for (String xpath_expression : argslist) {
            System.out.println(xpath_expression);
            //Remove brackets from attribute comparison
            for( Pattern brack_sub : ATTR_SUBST) {
                Matcher matcher = brack_sub.matcher(xpath_expression);
                boolean matchFound = matcher.find();

                if (matchFound) {
                    for (int i=1; i<=matcher.groupCount(); i++) {
                        String groupStr = matcher.group(i);
                        xpath_expression = matcher.replaceAll("/"+groupStr);
                    }
                }
            }

            lFullPaths.add(xpath_expression);
        }

        return lFullPaths;
    }
}