package gov.fnal.xpathsearch.cache;

import gov.fnal.xpath2ldap.util.Configuration;

import org.apache.commons.cli.Options;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.CommandLine;


public class XPathCacheLoadOptions {
   public static final String H = "H";
   public static final String B = "b";
   //public static final String USE_LDAP_AUX = "use_ldap_aux";
   public static final String CACHE_TYPE = "cache_type";

    private static Options options;
    public static String version = "put_maven_version_filter_here";
    public static String syntax = "XPathCacheLoad [options] host_url:port [--cache_type <xmldb|filesystem>] xpathexpression1 [xpathexpression2] ...";

    public static String header = "Available options:";

    public static String footer = "\n" +
            "\nJava properties:\n" +
            "-Dxpathsearch.home=full path of XPathCacheLoad installation directory\n" +
            "\nTry 'XPathCacheLoad <command> --help' for help on a specific command.\n" +
            "For example 'XPathCacheLoad --help'.\n" +
            "\n" +
            "LDAPsearch -- Search of LDAP data.\n" +
            "For additional information, see http://cd-97177.fnal.gov:2088/redmine\n" +
            "Bug Reports to <xpathsearch@gmail.com>";

    public static CommandLine processArguments(String[] args) {
        return Configuration.processArguments(args, getOptions(), version, syntax, header, footer);
    }

    public static Options getOptions() {
        if(options==null) setupOptions();
        return options;
    }

    private static void setupOptions() {
        // Read in extra properties from the conf subdirectory of the install directory.
        /*
        try {
            URL base_conf_url = Configuration.getModulePropertiesURL("xquerysearch.home");
            URL[] conf_urls = new URL[]{base_conf_url};
            Configuration.processModuleProperties("xquerysearch.properties", conf_urls);
        } catch (
                IOException ioe) {
            System.err.println("Warning: could not process properites file");
        }
        */
        // create the Options
        options = new Options();
        options.addOption("v", "version", false, "");
        options.addOption("h", "help", false, "");
        Option host   = OptionBuilder.hasArg()
                .withArgName("PROTOCOL://HOST:PORT")
                .withDescription(  "host_uri for search. Example: -H ldap://iu.grid.ui.edu:2710" )
                .create( H );
        options.addOption(host);
        Option base   = OptionBuilder.hasArg()
                .withArgName("o=BASE")
                .withDescription(  "base for LDAP search. Example: -b o=grid" )
                .create( B );
        options.addOption(base);
        //options.addOption("u", USE_LDAP_AUX, false, "Use auxilary LDAP classes");
        Option cache_type   = OptionBuilder.hasArg()
                .withArgName("Cache Type")
                .withDescription(  "Type of cache to use to save LDAP info, 'xmldb' or 'filesystem'. No caching if not specified." )
                .create( CACHE_TYPE );
        options.addOption(cache_type);
        options.addOption("u", "update", false, "Update the cache. Default is to create a new cache.");
        options.addOption("t", "test", false, "Run the unit tests");
    }
}